import pytest

from diofant import (I, O, Rational, Symbol, atanh, conjugate, elliptic_e,
                     elliptic_f, elliptic_k, elliptic_pi, gamma, hyper,
                     meijerg, oo, pi, sin, sqrt, tan, zoo)
from diofant.abc import m, n, z
from diofant.core.function import ArgumentIndexError
from diofant.utilities.randtest import random_complex_number as randcplx
from diofant.utilities.randtest import verify_derivative_numerically as td
from diofant.utilities.randtest import verify_numerically as tn


__all__ = ()

i = Symbol('i', integer=True)
j = Symbol('k', integer=True, positive=True)


def test_elliptic_k():
    assert elliptic_k(0) == pi/2
    assert elliptic_k(Rational(1, 2)) == 8*pi**Rational(3, 2)/gamma(-Rational(1, 4))**2
    assert elliptic_k(1) == zoo
    assert elliptic_k(-1) == gamma(Rational(1, 4))**2/(4*sqrt(2*pi))
    assert elliptic_k(oo) == 0
    assert elliptic_k(-oo) == 0
    assert elliptic_k(I*oo) == 0
    assert elliptic_k(-I*oo) == 0
    assert elliptic_k(zoo) == 0

    assert elliptic_k(z).diff(z) == (elliptic_e(z) - (1 - z)*elliptic_k(z))/(2*z*(1 - z))
    assert td(elliptic_k(z), z)
    pytest.raises(ArgumentIndexError, lambda: elliptic_k(z).fdiff(2))

    zi = Symbol('z', extended_real=False)
    assert elliptic_k(zi).conjugate() == elliptic_k(zi.conjugate())
    zr = Symbol('z', extended_real=True, negative=True)
    assert elliptic_k(zr).conjugate() == elliptic_k(zr)
    assert elliptic_k(z).conjugate() == conjugate(elliptic_k(z), evaluate=False)

    assert elliptic_k(z).rewrite(hyper) == \
        (pi/2)*hyper((Rational(1, 2), Rational(1, 2)), (1,), z)
    assert tn(elliptic_k(z), (pi/2)*hyper((Rational(1, 2), Rational(1, 2)), (1,), z))
    assert elliptic_k(z).rewrite(meijerg) == \
        meijerg(((Rational(1, 2), Rational(1, 2)), []), ((0,), (0,)), -z)/2
    assert tn(elliptic_k(z), meijerg(((Rational(1, 2), Rational(1, 2)), []), ((0,), (0,)), -z)/2)

    assert elliptic_k(z).series(z) == pi/2 + pi*z/8 + 9*pi*z**2/128 + \
        25*pi*z**3/512 + 1225*pi*z**4/32768 + 3969*pi*z**5/131072 + O(z**6)


def test_elliptic_f():
    assert elliptic_f(z, 0) == z
    assert elliptic_f(0, m) == 0
    assert elliptic_f(pi*i/2, m) == i*elliptic_k(m)
    assert elliptic_f(z, oo) == 0
    assert elliptic_f(z, -oo) == 0

    assert elliptic_f(-z, m) == -elliptic_f(z, m)

    assert elliptic_f(z, m).diff(z) == 1/sqrt(1 - m*sin(z)**2)
    assert elliptic_f(z, m).diff(m) == elliptic_e(z, m)/(2*m*(1 - m)) - elliptic_f(z, m)/(2*m) - \
        sin(2*z)/(4*(1 - m)*sqrt(1 - m*sin(z)**2))
    r = randcplx()
    assert td(elliptic_f(z, r), z)
    assert td(elliptic_f(r, m), m)
    pytest.raises(ArgumentIndexError, lambda: elliptic_f(z, m).fdiff(3))

    mi = Symbol('m', extended_real=False)
    assert elliptic_f(z, mi).conjugate() == elliptic_f(z.conjugate(), mi.conjugate())
    mr = Symbol('m', extended_real=True, negative=True)
    assert elliptic_f(z, mr).conjugate() == elliptic_f(z.conjugate(), mr)
    assert elliptic_f(z, m).conjugate() == conjugate(elliptic_f(z, m), evaluate=False)

    assert elliptic_f(z, m).series(z) == \
        z + z**5*(3*m**2/40 - m/30) + m*z**3/6 + O(z**6)


def test_elliptic_e():
    assert elliptic_e(z, 0) == z
    assert elliptic_e(0, m) == 0
    assert elliptic_e(i*pi/2, m) == i*elliptic_e(m)
    assert elliptic_e(z, oo) == zoo
    assert elliptic_e(z, -oo) == zoo
    assert elliptic_e(0) == pi/2
    assert elliptic_e(1) == 1
    assert elliptic_e(oo) == I*oo
    assert elliptic_e(-oo) == oo
    assert elliptic_e(zoo) == zoo

    assert elliptic_e(-z, m) == -elliptic_e(z, m)

    assert elliptic_e(z, m).diff(z) == sqrt(1 - m*sin(z)**2)
    assert elliptic_e(z, m).diff(m) == (elliptic_e(z, m) - elliptic_f(z, m))/(2*m)
    assert elliptic_e(z).diff(z) == (elliptic_e(z) - elliptic_k(z))/(2*z)
    r = randcplx()
    assert td(elliptic_e(r, m), m)
    assert td(elliptic_e(z, r), z)
    assert td(elliptic_e(z), z)
    pytest.raises(ArgumentIndexError, lambda: elliptic_e(z, m).fdiff(3))
    pytest.raises(ArgumentIndexError, lambda: elliptic_e(z).fdiff(2))

    mi = Symbol('m', extended_real=False)
    assert elliptic_e(z, mi).conjugate() == elliptic_e(z.conjugate(), mi.conjugate())
    assert elliptic_e(mi).conjugate() == elliptic_e(mi.conjugate())
    mr = Symbol('m', extended_real=True, negative=True)
    assert elliptic_e(z, mr).conjugate() == elliptic_e(z.conjugate(), mr)
    assert elliptic_e(mr).conjugate() == elliptic_e(mr)
    assert elliptic_e(z, m).conjugate() == conjugate(elliptic_e(z, m))
    assert elliptic_e(z).conjugate() == conjugate(elliptic_e(z))

    assert elliptic_e(z).rewrite(hyper) == (pi/2)*hyper((Rational(-1, 2), Rational(1, 2)), (1,), z)
    assert elliptic_e(z, m).rewrite(hyper) == elliptic_e(z, m)
    assert tn(elliptic_e(z), (pi/2)*hyper((Rational(-1, 2), Rational(1, 2)), (1,), z))
    assert elliptic_e(z).rewrite(meijerg) == \
        -meijerg(((Rational(1, 2), Rational(3, 2)), []), ((0,), (0,)), -z)/4
    assert elliptic_e(z, m).rewrite(meijerg) == elliptic_e(z, m)
    assert tn(elliptic_e(z), -meijerg(((Rational(1, 2), Rational(3, 2)), []), ((0,), (0,)), -z)/4)

    assert elliptic_e(z, m).series(z) == \
        z + z**5*(-m**2/40 + m/30) - m*z**3/6 + O(z**6)
    assert elliptic_e(z).series(z) == pi/2 - pi*z/8 - 3*pi*z**2/128 - \
        5*pi*z**3/512 - 175*pi*z**4/32768 - 441*pi*z**5/131072 + O(z**6)


def test_elliptic_pi():
    assert elliptic_pi(0, z, m) == elliptic_f(z, m)
    assert elliptic_pi(1, z, m) == elliptic_f(z, m) + \
        (sqrt(1 - m*sin(z)**2)*tan(z) - elliptic_e(z, m))/(1 - m)
    assert elliptic_pi(n, i*pi/2, m) == i*elliptic_pi(n, m)
    assert elliptic_pi(n, z, 0) == atanh(sqrt(n - 1)*tan(z))/sqrt(n - 1)
    assert elliptic_pi(n, z, n) == elliptic_f(z, n) - elliptic_pi(1, z, n) + tan(z)/sqrt(1 - n*sin(z)**2)
    assert elliptic_pi(oo, z, m) == 0
    assert elliptic_pi(-oo, z, m) == 0
    assert elliptic_pi(n, z, oo) == 0
    assert elliptic_pi(n, z, -oo) == 0
    assert elliptic_pi(0, m) == elliptic_k(m)
    assert elliptic_pi(1, m) == zoo
    assert elliptic_pi(n, 0) == pi/(2*sqrt(1 - n))
    assert elliptic_pi(2, 1) == -oo
    assert elliptic_pi(-1, 1) == oo
    assert elliptic_pi(n, n) == elliptic_e(n)/(1 - n)
    assert elliptic_pi(oo, m) == 0
    assert elliptic_pi(n, oo) == 0

    assert elliptic_pi(n, -z, m) == -elliptic_pi(n, z, m)

    ni, mi = Symbol('n', extended_real=False), Symbol('m', extended_real=False)
    assert elliptic_pi(ni, z, mi).conjugate() == \
        elliptic_pi(ni.conjugate(), z.conjugate(), mi.conjugate())
    nr, mr = Symbol('n', extended_real=True, negative=True), \
        Symbol('m', extended_real=True, negative=True)
    assert elliptic_pi(nr, z, mr).conjugate() == elliptic_pi(nr, z.conjugate(), mr)
    assert elliptic_pi(n, m).conjugate() == elliptic_pi(n.conjugate(), m.conjugate())
    assert elliptic_pi(n, z, m).conjugate() == conjugate(elliptic_pi(n, z, m))

    assert elliptic_pi(n, z, m).diff(n) == (elliptic_e(z, m) + (m - n)*elliptic_f(z, m)/n +
                                            (n**2 - m)*elliptic_pi(n, z, m)/n - n*sqrt(1 -
                                                                                       m*sin(z)**2)*sin(2*z)/(2*(1 - n*sin(z)**2)))/(2*(m - n)*(n - 1))
    assert elliptic_pi(n, z, m).diff(z) == 1/(sqrt(1 - m*sin(z)**2)*(1 - n*sin(z)**2))
    assert elliptic_pi(n, z, m).diff(m) == (elliptic_e(z, m)/(m - 1) + elliptic_pi(n, z, m) -
                                            m*sin(2*z)/(2*(m - 1)*sqrt(1 - m*sin(z)**2)))/(2*(n - m))
    assert elliptic_pi(n, m).diff(n) == (elliptic_e(m) + (m - n)*elliptic_k(m)/n +
                                         (n**2 - m)*elliptic_pi(n, m)/n)/(2*(m - n)*(n - 1))
    assert elliptic_pi(n, m).diff(m) == (elliptic_e(m)/(m - 1) + elliptic_pi(n, m))/(2*(n - m))
    rx, ry = randcplx(), randcplx()
    assert td(elliptic_pi(n, rx, ry), n)
    assert td(elliptic_pi(rx, z, ry), z)
    assert td(elliptic_pi(rx, ry, m), m)
    pytest.raises(ArgumentIndexError, lambda: elliptic_pi(n, z, m).fdiff(4))
    pytest.raises(ArgumentIndexError, lambda: elliptic_pi(n, m).fdiff(3))

    assert elliptic_pi(n, z, m).series(z) == z + z**3*(m/6 + n/3) + \
        z**5*(3*m**2/40 + m*n/10 - m/30 + n**2/5 - n/15) + O(z**6)



r"""Module that defines indexed objects

The classes IndexedBase, Indexed and Idx would represent a matrix element
M[i, j] as in the following graph::

    1) The Indexed class represents the entire indexed object.
              |
           ___|___
          '       '
           M[i, j]
          /   \__\______
          |             |
          |             |
          |     2) The Idx class represent indices and each Idx can
          |        optionally contain information about its range.
          |
    3) IndexedBase represents the `stem' of an indexed object, here `M'.
    The stem used by itself is usually taken to represent the entire array.

There can be any number of indices on an Indexed object.  No
transformation properties are implemented in these Base objects, but
implicit contraction of repeated indices is supported.

Note that the support for complicated (i.e. non-atomic) integer
expressions as indices is limited.  (This should be improved in
future releases.)

Examples
========

To express the above matrix element example you would write:

>>> M = IndexedBase('M')
>>> i, j = symbols('i j', cls=Idx)
>>> M[i, j]
M[i, j]

Repeated indices in a product implies a summation, so to express a
matrix-vector product in terms of Indexed objects:

>>> x = IndexedBase('x')
>>> M[i, j]*x[j]
x[j]*M[i, j]

If the indexed objects will be converted to component based arrays, e.g.
with the code printers or the autowrap framework, you also need to provide
(symbolic or numerical) dimensions.  This can be done by passing an
optional shape parameter to IndexedBase upon construction:

>>> dim1, dim2 = symbols('dim1 dim2', integer=True)
>>> A = IndexedBase('A', shape=(dim1, 2*dim1, dim2))
>>> A.shape
(dim1, 2*dim1, dim2)
>>> A[i, j, 3].shape
(dim1, 2*dim1, dim2)

If an IndexedBase object has no shape information, it is assumed that the
array is as large as the ranges of its indices:

>>> i = Idx('i', m)
>>> j = Idx('j', n)
>>> M[i, j].shape
(m, n)
>>> M[i, j].ranges
[(0, m - 1), (0, n - 1)]

The above can be compared with the following:

>>> A[i, 2, j].shape
(dim1, 2*dim1, dim2)
>>> A[i, 2, j].ranges
[(0, m - 1), None, (0, n - 1)]

To analyze the structure of indexed expressions, you can use the methods
get_indices() and get_contraction_structure():

>>> get_indices(A[i, j, j])
({i}, {})
>>> get_contraction_structure(A[i, j, j])
{(j,): {A[i, j, j]}}

See the appropriate docstrings for a detailed explanation of the output.

"""

#   TODO:  (some ideas for improvement)
#
#   o test and guarantee numpy compatibility
#      - implement full support for broadcasting
#      - strided arrays
#
#   o more functions to analyze indexed expressions
#      - identify standard constructs, e.g matrix-vector product in a subexpression
#
#   o functions to generate component based arrays (numpy and diofant.Matrix)
#      - generate a single array directly from Indexed
#      - convert simple sub-expressions
#
#   o sophisticated indexing (possibly in subclasses to preserve simplicity)
#      - Idx with range smaller than dimension of Indexed
#      - Idx with stepsize != 1
#      - Idx with step determined by function call

from ..core import Dummy, Expr, Symbol, Tuple, oo, sympify
from ..core.compatibility import NotIterable, is_sequence


class IndexException(Exception):
    """Generic index error."""

    pass


class Indexed(Expr):
    """Represents a mathematical object with indices.

    >>> i, j = symbols('i j', cls=Idx)
    >>> Indexed('A', i, j)
    A[i, j]

    It is recommended that Indexed objects are created via IndexedBase:

    >>> A = IndexedBase('A')
    >>> Indexed('A', i, j) == A[i, j]
    True

    """

    is_commutative = True

    def __new__(cls, base, *args, **kw_args):
        from ..utilities import filldedent

        from .array.ndim_array import NDimArray
        from ..matrices.matrices import MatrixBase

        if not args:
            raise IndexException('Indexed needs at least one index.')
        if isinstance(base, (str, Symbol)):
            base = IndexedBase(base)
        elif not hasattr(base, '__getitem__') and not isinstance(base, IndexedBase):
            raise TypeError(filldedent("""
                Indexed expects string, Symbol or IndexedBase as base."""))
        args = list(map(sympify, args))
        if isinstance(base, (NDimArray, Tuple, MatrixBase)) and all(i.is_number for i in args):
            return base[args]
        return Expr.__new__(cls, base, *args, **kw_args)

    @property
    def base(self):
        """Returns the IndexedBase of the Indexed object.

        Examples
        ========

        >>> i, j = symbols('i j', cls=Idx)
        >>> Indexed('A', i, j).base
        A
        >>> B = IndexedBase('B')
        >>> B == B[i, j].base
        True

        """
        return self.args[0]

    @property
    def indices(self):
        """
        Returns the indices of the Indexed object.

        Examples
        ========

        >>> i, j = symbols('i j', cls=Idx)
        >>> Indexed('A', i, j).indices
        (i, j)

        """
        return self.args[1:]

    @property
    def rank(self):
        """
        Returns the rank of the Indexed object.

        Examples
        ========

        >>> i, j, k, l, m = symbols('i:m', cls=Idx)
        >>> Indexed('A', i, j).rank
        2
        >>> q = Indexed('A', i, j, k, l, m)
        >>> q.rank
        5
        >>> q.rank == len(q.indices)
        True

        """
        return len(self.args) - 1

    @property
    def shape(self):
        """Returns a list with dimensions of each index.

        Dimensions is a property of the array, not of the indices.  Still, if
        the IndexedBase does not define a shape attribute, it is assumed that
        the ranges of the indices correspond to the shape of the array.

        >>> i = Idx('i', m)
        >>> j = Idx('j', m)
        >>> A = IndexedBase('A', shape=(n, n))
        >>> B = IndexedBase('B')
        >>> A[i, j].shape
        (n, n)
        >>> B[i, j].shape
        (m, m)

        """
        from ..utilities import filldedent

        if self.base.shape:
            return self.base.shape
        try:
            return Tuple(*[i.upper - i.lower + 1 for i in self.indices])
        except AttributeError:
            raise IndexException(filldedent(f"""
                Range is not defined for all indices in: {self}"""))
        except TypeError:
            raise IndexException(filldedent(f"""
                Shape cannot be inferred from Idx with
                undefined range: {self}"""))

    @property
    def ranges(self):
        """Returns a list of tuples with lower and upper range of each index.

        If an index does not define the data members upper and lower, the
        corresponding slot in the list contains ``None`` instead of a tuple.

        Examples
        ========

        >>> Indexed('A', Idx('i', 2), Idx('j', 4), Idx('k', 8)).ranges
        [(0, 1), (0, 3), (0, 7)]
        >>> Indexed('A', Idx('i', 3), Idx('j', 3), Idx('k', 3)).ranges
        [(0, 2), (0, 2), (0, 2)]
        >>> Indexed('A', x, y, z).ranges
        [None, None, None]

        """
        ranges = []
        for i in self.indices:
            try:
                ranges.append(Tuple(i.lower, i.upper))
            except AttributeError:
                ranges.append(None)
        return ranges

    def _diofantstr(self, p):
        indices = list(map(p.doprint, self.indices))
        return f"{p.doprint(self.base)}[{', '.join(indices)}]"


class IndexedBase(Expr, NotIterable):
    """Represent the base or stem of an indexed object

    The IndexedBase class represent an array that contains elements. The main purpose
    of this class is to allow the convenient creation of objects of the Indexed
    class.  The __getitem__ method of IndexedBase returns an instance of
    Indexed.  Alone, without indices, the IndexedBase class can be used as a
    notation for e.g. matrix equations, resembling what you could do with the
    Symbol class.  But, the IndexedBase class adds functionality that is not
    available for Symbol instances:

      -  An IndexedBase object can optionally store shape information.  This can
         be used in to check array conformance and conditions for numpy
         broadcasting.  (TODO)
      -  An IndexedBase object implements syntactic sugar that allows easy symbolic
         representation of array operations, using implicit summation of
         repeated indices.
      -  The IndexedBase object symbolizes a mathematical structure equivalent
         to arrays, and is recognized as such for code generation and automatic
         compilation and wrapping.

    >>> A = IndexedBase('A')
    >>> A
    A
    >>> type(A)
    <class 'diofant.tensor.indexed.IndexedBase'>

    When an IndexedBase object receives indices, it returns an array with named
    axes, represented by an Indexed object:

    >>> i, j = symbols('i j', integer=True)
    >>> A[i, j, 2]
    A[i, j, 2]
    >>> type(A[i, j, 2])
    <class 'diofant.tensor.indexed.Indexed'>

    The IndexedBase constructor takes an optional shape argument.  If given,
    it overrides any shape information in the indices. (But not the index
    ranges!)

    >>> o, p = symbols('o p', integer=True)
    >>> i = Idx('i', m)
    >>> j = Idx('j', n)
    >>> A[i, j].shape
    (m, n)
    >>> B = IndexedBase('B', shape=(o, p))
    >>> B[i, j].shape
    (o, p)

    """

    def __new__(cls, label, shape=None, **kw_args):
        if isinstance(label, str):
            label = Symbol(label)
        elif isinstance(label, (Dummy, Symbol)):
            pass
        else:
            label = sympify(label, strict=True)

        obj = Expr.__new__(cls, label, **kw_args)
        if is_sequence(shape):
            obj._shape = Tuple(*shape)
        elif shape is not None:
            obj._shape = Tuple(shape)
        else:
            obj._shape = shape
        return obj

    @property
    def args(self):
        """Returns the arguments used to create this IndexedBase object.

        Examples
        ========

        >>> IndexedBase('A', shape=(x, y)).args
        (A, (x, y))

        """
        if self._shape:
            return super().args + (self._shape,)
        else:
            return super().args

    def _hashable_content(self):
        return Expr._hashable_content(self) + (self._shape,)

    def __getitem__(self, indices, **kw_args):
        if is_sequence(indices):
            # Special case needed because M[*my_tuple] is a syntax error.
            if self.shape and len(self.shape) != len(indices):
                raise IndexException('Rank mismatch.')
            return Indexed(self, *indices, **kw_args)
        else:
            if self.shape and len(self.shape) != 1:
                raise IndexException('Rank mismatch.')
            return Indexed(self, indices, **kw_args)

    @property
    def shape(self):
        """Returns the shape of the IndexedBase object.

        Examples
        ========

        >>> IndexedBase('A', shape=(x, y)).shape
        (x, y)

        Note: If the shape of the IndexedBase is specified, it will override
        any shape information given by the indices.

        >>> A = IndexedBase('A', shape=(x, y))
        >>> B = IndexedBase('B')
        >>> i = Idx('i', 2)
        >>> j = Idx('j', 1)
        >>> A[i, j].shape
        (x, y)
        >>> B[i, j].shape
        (2, 1)

        """
        return self._shape

    @property
    def label(self):
        """Returns the label of the IndexedBase object.

        Examples
        ========

        >>> IndexedBase('A', shape=(x, y)).label
        A

        """
        return self.args[0]

    def _diofantstr(self, p):
        return p.doprint(self.label)


class Idx(Expr):
    """Represents an integer index as an Integer or integer expression.

    There are a number of ways to create an Idx object.  The constructor
    takes two arguments:

    ``label``
        An integer or a symbol that labels the index.
    ``range``
        Optionally you can specify a range as either

    - Symbol or integer: This is interpreted as a dimension. Lower and
      upper bounds are set to 0 and range - 1, respectively.
    - tuple: The two elements are interpreted as the lower and upper
      bounds of the range, respectively.

    Note: the Idx constructor is rather pedantic in that it only accepts
    integer arguments.  The only exception is that you can use oo and -oo to
    specify an unbounded range.  For all other cases, both label and bounds
    must be declared as integers, e.g. if n is given as an argument then
    n.is_integer must return True.

    For convenience, if the label is given as a string it is automatically
    converted to an integer symbol.  (Note: this conversion is not done for
    range or dimension arguments.)

    Examples
    ========

    >>> i, L, U = symbols('i L U', integer=True)

    If a string is given for the label an integer Symbol is created and the
    bounds are both None:

    >>> idx = Idx('qwerty')
    >>> idx
    qwerty
    >>> idx.lower, idx.upper
    (None, None)

    Both upper and lower bounds can be specified:

    >>> idx = Idx(i, (L, U))
    >>> idx
    i
    >>> idx.lower, idx.upper
    (L, U)

    When only a single bound is given it is interpreted as the dimension
    and the lower bound defaults to 0:

    >>> idx = Idx(i, n)
    >>> idx.lower, idx.upper
    (0, n - 1)
    >>> idx = Idx(i, 4)
    >>> idx.lower, idx.upper
    (0, 3)
    >>> idx = Idx(i, oo)
    >>> idx.lower, idx.upper
    (0, oo)

    The label can be a literal integer instead of a string/Symbol:

    >>> idx = Idx(2, n)
    >>> idx.lower, idx.upper
    (0, n - 1)
    >>> idx.label
    2

    """

    is_integer = True

    def __new__(cls, label, range=None, **kw_args):
        from ..utilities import filldedent

        if isinstance(label, str):
            label = Symbol(label, integer=True)
        label, range = list(map(sympify, (label, range)))

        if not label.is_integer:
            raise TypeError('Idx object requires an integer label.')

        elif is_sequence(range):
            if len(range) != 2:
                raise ValueError(filldedent(f"""
                    Idx range tuple must have length 2, but got {len(range)}"""))
            for bound in range:
                if not (bound.is_integer or abs(bound) is oo):
                    raise TypeError('Idx object requires integer bounds.')
            args = label, Tuple(*range)
        elif isinstance(range, Expr):
            if not (range.is_integer or range is oo):
                raise TypeError('Idx object requires an integer dimension.')
            args = label, Tuple(0, range - 1)
        elif range:
            raise TypeError(filldedent("""
                The range must be an ordered iterable or
                integer Diofant expression."""))
        else:
            args = label,

        obj = Expr.__new__(cls, *args, **kw_args)
        return obj

    @property
    def label(self):
        """Returns the label (Integer or integer expression) of the Idx object.

        Examples
        ========

        >>> Idx(2).label
        2
        >>> j = Symbol('j', integer=True)
        >>> Idx(j).label
        j
        >>> Idx(j + 1).label
        j + 1

        """
        return self.args[0]

    @property
    def lower(self):
        """Returns the lower bound of the Index.

        Examples
        ========

        >>> Idx('j', 2).lower
        0
        >>> Idx('j', 5).lower
        0
        >>> Idx('j').lower is None
        True

        """
        try:
            return self.args[1][0]
        except IndexError:
            return

    @property
    def upper(self):
        """Returns the upper bound of the Index.

        Examples
        ========

        >>> Idx('j', 2).upper
        1
        >>> Idx('j', 5).upper
        4
        >>> Idx('j').upper is None
        True

        """
        try:
            return self.args[1][1]
        except IndexError:
            return

    def _diofantstr(self, p):
        return p.doprint(self.label)



"""Special sets."""

from ..core import Basic, Expr, Integer, Lambda, Rational, S, oo
from ..core.compatibility import as_int
from ..core.singleton import Singleton
from ..core.sympify import converter, sympify
from ..logic import false, true
from ..utilities.iterables import cantor_product
from .sets import EmptySet, FiniteSet, Intersection, Interval, Set


class Naturals(Set, metaclass=Singleton):
    """The set of natural numbers.

    Represents the natural numbers (or counting numbers) which are all
    positive integers starting from 1. This set is also available as
    the Singleton, S.Naturals.

    Examples
    ========

    >>> 5 in S.Naturals
    True
    >>> iterable = iter(S.Naturals)
    >>> next(iterable)
    1
    >>> next(iterable)
    2
    >>> next(iterable)
    3
    >>> S.Naturals.intersection(Interval(0, 10))
    Range(1, 11, 1)

    See Also
    ========

    Naturals0 : non-negative integers
    Integers : also includes negative integers

    """

    is_iterable = True
    _inf = Integer(1)
    _sup = oo

    def _intersection(self, other):
        if other.is_Interval:
            return Intersection(
                S.Integers, other, Interval(self._inf, oo, False, True))

    def _contains(self, other):
        if not isinstance(other, Expr):
            return false
        elif other.is_positive and other.is_integer:
            return true
        elif other.is_integer is False or other.is_positive is False:
            return false

    def __iter__(self):
        i = self._inf
        while True:
            yield i
            i = i + 1

    @property
    def _boundary(self):
        return self


class Naturals0(Naturals):
    """The set of natural numbers, starting from 0.

    Represents the whole numbers which are all the non-negative
    integers, inclusive of zero.

    See Also
    ========

    Naturals : positive integers
    Integers : also includes the negative integers

    """

    _inf = Integer(0)

    def _contains(self, other):
        if not isinstance(other, Expr):
            return false
        elif other.is_integer and other.is_nonnegative:
            return true
        elif other.is_integer is False or other.is_nonnegative is False:
            return false


class Integers(Set, metaclass=Singleton):
    """The set of all integers.

    Represents all integers: positive, negative and zero. This set
    is also available as the Singleton, S.Integers.

    Examples
    ========

    >>> 5 in S.Naturals
    True
    >>> iterable = iter(S.Integers)
    >>> next(iterable)
    0
    >>> next(iterable)
    1
    >>> next(iterable)
    -1
    >>> next(iterable)
    2

    >>> S.Integers.intersection(Interval(-4, 4))
    Range(-4, 5, 1)

    See Also
    ========

    Naturals0 : non-negative integers
    Integers : positive and negative integers and zero

    """

    is_iterable = True

    def _intersection(self, other):
        from ..functions import floor, ceiling
        if other is Interval(-oo, oo, True, True) or other is S.Reals:
            return self
        elif other.is_Interval:
            s = Range(ceiling(other.left), floor(other.right) + 1)
            return s.intersection(other)  # take out endpoints if open interval

    def _contains(self, other):
        if not isinstance(other, Expr):
            return false
        elif other.is_integer:
            return true
        elif other.is_integer is False:
            return false

    def __iter__(self):
        yield Integer(0)
        i = Integer(1)
        while True:
            yield i
            yield -i
            i = i + 1

    @property
    def _inf(self):
        return -oo

    @property
    def _sup(self):
        return oo

    @property
    def _boundary(self):
        return self

    def _eval_imageset(self, f):
        from ..core import Wild
        expr = f.expr
        if len(f.variables) > 1:
            return
        n = f.variables[0]

        a = Wild('a')
        b = Wild('b')

        match = expr.match(a*n + b)
        if match[a].is_negative:
            expr = -expr

        match = expr.match(a*n + b)
        if match[a] == 1 and match[b].is_integer:
            expr = expr - match[b]

        return ImageSet(Lambda(n, expr), S.Integers)


class Rationals(Set, metaclass=Singleton):
    """The set of all rationals."""

    def _contains(self, other):
        if other.is_rational:
            return true
        elif other.is_rational is False:
            return false

    @property
    def _inf(self):
        return -oo

    @property
    def _sup(self):
        return oo

    @property
    def _boundary(self):
        return self

    def __iter__(self):
        seen = []
        pairs = cantor_product(S.Integers, S.Naturals)
        while True:
            n, d = next(pairs)
            r = Rational(n, d)
            if r not in seen:
                seen.append(r)
                yield r


class Reals(Interval, metaclass=Singleton):
    """The set of all reals."""

    def __new__(cls):
        return Interval.__new__(cls, -oo, oo, True, True)


class ImageSet(Set):
    """Image of a set under a mathematical function.

    Examples
    ========

    >>> squares = ImageSet(Lambda(x, x**2), S.Naturals)
    >>> 4 in squares
    True
    >>> 5 in squares
    False

    >>> FiniteSet(0, 1, 2, 3, 4, 5, 6, 7, 9, 10).intersection(squares)
    {1, 4, 9}

    >>> square_iterable = iter(squares)
    >>> for i in range(4):
    ...     next(square_iterable)
    1
    4
    9
    16

    If you want to get value for `x` = 2, 1/2 etc. (Please check whether the
    `x` value is in `base_set` or not before passing it as args)

    >>> squares.lamda(2)
    4
    >>> squares.lamda(Rational(1, 2))
    1/4

    """

    def __new__(cls, lamda, base_set):
        return Basic.__new__(cls, lamda, base_set)

    lamda = property(lambda self: self.args[0])
    base_set = property(lambda self: self.args[1])

    def __iter__(self):
        already_seen = set()
        for i in self.base_set:
            val = self.lamda(i)
            if val in already_seen:
                continue
            else:
                already_seen.add(val)
                yield val

    def _contains(self, other):
        from ..solvers import solve

        L = self.lamda
        if len(self.lamda.variables) > 1:
            return  # pragma: no cover

        solns = solve(L.expr - other, L.variables[0])

        for soln in solns:
            if soln[L.variables[0]] in self.base_set:
                return true
        return false

    @property
    def is_iterable(self):
        return self.base_set.is_iterable

    def _intersection(self, other):
        from ..core import Dummy
        from ..solvers.diophantine import diophantine
        from .sets import imageset
        if self.base_set is S.Integers:
            if isinstance(other, ImageSet) and other.base_set is S.Integers:
                f, g = self.lamda.expr, other.lamda.expr
                n, m = self.lamda.variables[0], other.lamda.variables[0]

                # Diophantine sorts the solutions according to the alphabetic
                # order of the variable names, since the result should not depend
                # on the variable name, they are replaced by the dummy variables
                # below
                a, b = Dummy('a'), Dummy('b')
                f, g = f.subs({n: a}), g.subs({m: b})
                solns_set = diophantine(f - g)
                if solns_set == set():
                    return EmptySet()
                solns = list(diophantine(f - g))
                if len(solns) == 1:
                    t = list(solns[0][0].free_symbols)[0]
                else:
                    return  # pragma: no cover

                # since 'a' < 'b'
                return imageset(Lambda(t, f.subs({a: solns[0][0]})), S.Integers)

        if other == S.Reals:
            from ..solvers.diophantine import diophantine
            from ..core import expand_complex

            if len(self.lamda.variables) > 1 or self.base_set is not S.Integers:
                return  # pragma: no cover

            f = self.lamda.expr
            n = self.lamda.variables[0]

            n_ = Dummy(n.name, integer=True)
            f_ = f.subs({n: n_})

            re, im = map(expand_complex, f_.as_real_imag())

            sols = list(diophantine(im, n_))
            if all(s[0].has(n_) is False for s in sols):
                s = FiniteSet(*[s[0] for s in sols])
            elif len(sols) == 1 and sols[0][0].has(n_):
                s = imageset(Lambda(n_, sols[0][0]), S.Integers)
            else:
                return  # pragma: no cover

            return imageset(Lambda(n_, re), self.base_set.intersection(s))


class Range(Set):
    """Represents a range of integers.

    Examples
    ========

    >>> list(Range(5))
    [0, 1, 2, 3, 4]
    >>> list(Range(10, 15))
    [10, 11, 12, 13, 14]
    >>> list(Range(10, 20, 2))
    [10, 12, 14, 16, 18]
    >>> list(Range(20, 10, -2))
    [12, 14, 16, 18, 20]

    """

    is_iterable = True

    def __new__(cls, *args):
        from ..functions import ceiling
        if len(args) == 1 and isinstance(args[0], range):
            args = args[0].start, args[0].stop, args[0].step

        # expand range
        slc = slice(*args)
        start, stop, step = slc.start or 0, slc.stop, slc.step or 1
        try:
            start, stop, step = [w if w in [-oo, oo] else Integer(as_int(w))
                                 for w in (start, stop, step)]
        except ValueError:
            raise ValueError('Inputs to Range must be Integer Valued\n' +
                             'Use ImageSets of Ranges for other cases')

        if not step.is_finite:
            raise ValueError('Infinite step is not allowed')
        if start == stop:
            return S.EmptySet

        n = ceiling((stop - start)/step)
        if n <= 0:
            return S.EmptySet

        # normalize args: regardless of how they are entered they will show
        # canonically as Range(inf, sup, step) with step > 0
        if n.is_finite:
            start, stop = sorted((start, start + (n - 1)*step))
        else:
            start, stop = sorted((start, stop - step))

        step = abs(step)
        if (start, stop) == (-oo, oo):
            raise ValueError('Both the start and end value of '
                             'Range cannot be unbounded')
        else:
            return Basic.__new__(cls, start, stop + step, step)

    start = property(lambda self: self.args[0])
    stop = property(lambda self: self.args[1])
    step = property(lambda self: self.args[2])

    def _intersection(self, other):
        from ..functions import floor, ceiling, Min, Max
        if other.is_Interval:
            osup = other.sup
            oinf = other.inf
            # if other is [0, 10) we can only go up to 9
            if osup.is_integer and other.right_open:
                osup -= 1
            if oinf.is_integer and other.left_open:
                oinf += 1

            # Take the most restrictive of the bounds set by the two sets
            # round inwards
            inf = ceiling(Max(self.inf, oinf))
            sup = floor(Min(self.sup, osup))
            # if we are off the sequence, get back on
            if inf.is_finite and self.inf.is_finite:
                off = (inf - self.inf) % self.step
            else:
                off = Integer(0)
            if off:
                inf += self.step - off

            return Range(inf, sup + 1, self.step)

        if other == S.Naturals:
            return self._intersection(Interval(1, oo, False, True))

        if other == S.Integers:
            return self

    def _contains(self, other):
        if (((self.start - other)/self.step).is_integer or
                ((self.stop - other)/self.step).is_integer):
            return sympify(other >= self.inf and other <= self.sup, strict=True)
        else:
            return false

    def __iter__(self):
        if self.start == -oo:
            i = self.stop - self.step
            step = -self.step
        else:
            i = self.start
            step = self.step

        while(i < self.stop and i >= self.start):
            yield i
            i += step

    def __len__(self):
        return int((self.stop - self.start)//self.step)

    def __bool__(self):
        return True

    def _ith_element(self, i):
        return self.start + i*self.step

    @property
    def _last_element(self):
        if self.stop is oo:
            return oo
        elif self.start == -oo:
            return self.stop - self.step
        else:
            return self._ith_element(len(self) - 1)

    @property
    def _inf(self):
        return self.start

    @property
    def _sup(self):
        return self.stop - self.step

    @property
    def _boundary(self):
        return self


converter[range] = Range



"""Symbolic primitives + unicode/ASCII abstraction for pretty.py"""

import unicodedata

from ...core.alphabets import greeks
from ..conventions import split_super_sub


# first, setup unicodedate environment
def U(name):
    """Unicode character by name or None if not found."""
    try:
        u = unicodedata.lookup(name)
    except KeyError:
        u = None

    return u


# prefix conventions when constructing tables
# L   - LATIN     i
# G   - GREEK     beta
# D   - DIGIT     0
# S   - SYMBOL    +


__all__ = ('greek_unicode', 'sub', 'sup', 'xsym', 'vobj', 'hobj',
           'pretty_symbol', 'annotated')


_use_unicode = False


def pretty_use_unicode(flag=None):
    """Set whether pretty-printer should use unicode by default."""
    global _use_unicode
    if flag is None:
        return _use_unicode

    use_unicode_prev = _use_unicode
    _use_unicode = flag
    return use_unicode_prev


# GREEK


def g(l):
    return U('GREEK SMALL LETTER %s' % l.upper())


def G(l):
    return U('GREEK CAPITAL LETTER %s' % l.upper())


greek_letters = list(greeks)  # make a copy
# deal with Unicode's funny spelling of lambda
greek_letters[greek_letters.index('lambda')] = 'lamda'

# {}  greek letter -> (g,G)
greek_unicode = {l: (g(l), G(l)) for l in greek_letters}
greek_unicode = {L: g(L) for L in greek_letters}
greek_unicode.update((L[0].upper() + L[1:], G(L)) for L in greek_letters)

# aliases
greek_unicode['lambda'] = greek_unicode['lamda']
greek_unicode['Lambda'] = greek_unicode['Lamda']
greek_unicode['varsigma'] = '\N{GREEK SMALL LETTER FINAL SIGMA}'

digit_2txt = {
    '0':    'ZERO',
    '1':    'ONE',
    '2':    'TWO',
    '3':    'THREE',
    '4':    'FOUR',
    '5':    'FIVE',
    '6':    'SIX',
    '7':    'SEVEN',
    '8':    'EIGHT',
    '9':    'NINE',
}

symb_2txt = {
    '+':    'PLUS SIGN',
    '-':    'MINUS',
    '=':    'EQUALS SIGN',
    '(':    'LEFT PARENTHESIS',
    ')':    'RIGHT PARENTHESIS',
    '[':    'LEFT SQUARE BRACKET',
    ']':    'RIGHT SQUARE BRACKET',
    '{':    'LEFT CURLY BRACKET',
    '}':    'RIGHT CURLY BRACKET',

    # non-std
    '{}':   'CURLY BRACKET',
    'sum':  'SUMMATION',
    'int':  'INTEGRAL',
}

# SUBSCRIPT & SUPERSCRIPT


def LSUB(letter):
    return U('LATIN SUBSCRIPT SMALL LETTER %s' % letter.upper())


def GSUB(letter):
    return U('GREEK SUBSCRIPT SMALL LETTER %s' % letter.upper())


def DSUB(digit):
    return U(f'SUBSCRIPT {digit_2txt[digit]}')


def SSUB(symb):
    return U(f'SUBSCRIPT {symb_2txt[symb]}')


def LSUP(letter):
    return U('SUPERSCRIPT LATIN SMALL LETTER %s' % letter.upper())


def DSUP(digit):
    return U(f'SUPERSCRIPT {digit_2txt[digit]}')


def SSUP(symb):
    return U(f'SUPERSCRIPT {symb_2txt[symb]}')


sub = {}    # symb -> subscript symbol
sup = {}    # symb -> superscript symbol

# latin subscripts
for l in 'aeioruvxhklmnpst':
    sub[l] = LSUB(l)

for l in 'in':
    sup[l] = LSUP(l)

for gl in ['beta', 'gamma', 'rho', 'phi', 'chi']:
    sub[gl] = GSUB(gl)

for d in [str(i) for i in range(10)]:
    sub[d] = DSUB(d)
    sup[d] = DSUP(d)

for s in '+-=()':
    sub[s] = SSUB(s)
    sup[s] = SSUP(s)

# Variable modifiers
# TODO: Is it worth trying to handle faces with, e.g., 'MATHEMATICAL BOLD CAPITAL A'?
# TODO: Make brackets adjust to height of contents
modifier_dict = {
    # Accents
    'mathring': lambda s: s + '\N{COMBINING RING ABOVE}',
    'ddddot': lambda s: s + '\N{COMBINING DIAERESIS}\N{COMBINING DIAERESIS}',
    'dddot': lambda s: s + '\N{COMBINING DIAERESIS}\N{COMBINING DOT ABOVE}',
    'ddot': lambda s: s + '\N{COMBINING DIAERESIS}',
    'dot': lambda s: s + '\N{COMBINING DOT ABOVE}',
    'check': lambda s: s + '\N{COMBINING CARON}',
    'breve': lambda s: s + '\N{COMBINING BREVE}',
    'acute': lambda s: s + '\N{COMBINING ACUTE ACCENT}',
    'grave': lambda s: s + '\N{COMBINING GRAVE ACCENT}',
    'tilde': lambda s: s + '\N{COMBINING TILDE}',
    'hat': lambda s: s + '\N{COMBINING CIRCUMFLEX ACCENT}',
    'bar': lambda s: s + '\N{COMBINING OVERLINE}',
    'vec': lambda s: s + '\N{COMBINING RIGHT ARROW ABOVE}',
    'prime': lambda s: s + '\N{PRIME}',
    'prm': lambda s: s + '\N{PRIME}',
    # # Faces -- these are here for some compatibility with latex printing
    # 'bold': lambda s: s,
    # 'bm': lambda s: s,
    # 'cal': lambda s: s,
    # 'scr': lambda s: s,
    # 'frak': lambda s: s,
    # Brackets
    'norm': lambda s: '\N{DOUBLE VERTICAL LINE}' + s + '\N{DOUBLE VERTICAL LINE}',
    'avg': lambda s: '\N{MATHEMATICAL LEFT ANGLE BRACKET}' + s + '\N{MATHEMATICAL RIGHT ANGLE BRACKET}',
    'abs': lambda s: '\N{VERTICAL LINE}' + s + '\N{VERTICAL LINE}',
    'mag': lambda s: '\N{VERTICAL LINE}' + s + '\N{VERTICAL LINE}',
}

# VERTICAL OBJECTS


def HUP(symb):
    return U(f'{symb_2txt[symb]} UPPER HOOK')


def CUP(symb):
    return U(f'{symb_2txt[symb]} UPPER CORNER')


def MID(symb):
    return U(f'{symb_2txt[symb]} MIDDLE PIECE')


def EXT(symb):
    return U(f'{symb_2txt[symb]} EXTENSION')


def HLO(symb):
    return U(f'{symb_2txt[symb]} LOWER HOOK')


def CLO(symb):
    return U(f'{symb_2txt[symb]} LOWER CORNER')


# {} '('  ->  (extension, start, end, middle) 1-character
_xobj_unicode = {

    # vertical symbols
    #       (( ext, top, bot, mid ), c1)
    '(':    (( EXT('('), HUP('('), HLO('(') ), '('),
    ')':    (( EXT(')'), HUP(')'), HLO(')') ), ')'),
    '[':    (( EXT('['), CUP('['), CLO('[') ), '['),
    ']':    (( EXT(']'), CUP(']'), CLO(']') ), ']'),
    '{':    (( EXT('{}'), HUP('{'), HLO('{'), MID('{') ), '{'),
    '}':    (( EXT('{}'), HUP('}'), HLO('}'), MID('}') ), '}'),
    '|':    '\N{BOX DRAWINGS LIGHT VERTICAL}',

    '<':    (('\N{BOX DRAWINGS LIGHT VERTICAL}',
              '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT}',
              '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT}'), '<'),

    '>':    (('\N{BOX DRAWINGS LIGHT VERTICAL}',
              '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT}',
              '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT}'), '>'),

    'lfloor': (( EXT('['), EXT('['), CLO('[') ), '\N{LEFT FLOOR}'),
    'rfloor': (( EXT(']'), EXT(']'), CLO(']') ), '\N{RIGHT FLOOR}'),
    'lceil':  (( EXT('['), CUP('['), EXT('[') ), '\N{LEFT CEILING}'),
    'rceil':  (( EXT(']'), CUP(']'), EXT(']') ), '\N{RIGHT CEILING}'),

    'int':  (( EXT('int'), '\N{TOP HALF INTEGRAL}', '\N{BOTTOM HALF INTEGRAL}' ), '\N{INTEGRAL}'),
    'sum':  (( '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT}', '_', '\N{OVERLINE}', '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT}'), '\N{N-ARY SUMMATION}'),

    # horizontal objects
    # '-':   '-',
    '-':    '\N{BOX DRAWINGS LIGHT HORIZONTAL}',
    '_':    '\N{LOW LINE}',
    # We used to use this, but LOW LINE looks better for roots, as it's a
    # little lower (i.e., it lines up with the / perfectly.  But perhaps this
    # one would still be wanted for some cases?
    # '_':    '\N{HORIZONTAL SCAN LINE-9}',

    # diagonal objects '\' & '/' ?
    '/':    '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT}',
    '\\':   '\N{BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT}',
}

_xobj_ascii = {
    # vertical symbols
    #       (( ext, top, bot, mid ), c1)
    '(':    (( '|', '/', '\\' ), '('),
    ')':    (( '|', '\\', '/' ), ')'),

    # XXX this looks ugly
    #   '[':    (( '|', '-', '-' ), '['),
    #   ']':    (( '|', '-', '-' ), ']'),
    # XXX not so ugly :(
    '[':    (( '[', '[', '[' ), '['),
    ']':    (( ']', ']', ']' ), ']'),

    '{':    (( '|', '/', '\\', '<' ), '{'),
    '}':    (( '|', '\\', '/', '>' ), '}'),
    '|':    '|',

    '<':    (( '|', '/', '\\' ), '<'),
    '>':    (( '|', '\\', '/' ), '>'),

    'int':  ( ' | ', '  /', '/  ' ),

    # horizontal objects
    '-':    '-',
    '_':    '_',

    # diagonal objects '\' & '/' ?
    '/':    '/',
    '\\':   '\\',
}


def xobj(symb, length):
    """Construct spatial object of given length.

    return: [] of equal-length strings
    """
    if length <= 0:
        raise ValueError('Length should be greater than 0')

    # TODO robustify when no unicodedat available
    if _use_unicode:
        _xobj = _xobj_unicode
    else:
        _xobj = _xobj_ascii

    vinfo = _xobj[symb]

    c1 = top = bot = mid = None

    if not isinstance(vinfo, tuple):        # 1 entry
        ext = vinfo
    else:
        if isinstance(vinfo[0], tuple):     # (vlong), c1
            vlong = vinfo[0]
            c1 = vinfo[1]
        else:                               # (vlong), c1
            vlong = vinfo

        ext = vlong[0]

        try:
            top = vlong[1]
            bot = vlong[2]
            mid = vlong[3]
        except IndexError:
            pass

    if c1 is None:
        c1 = ext
    if top is None:
        top = ext
    if bot is None:
        bot = ext
    if mid is not None:
        if (length % 2) == 0:
            # even height, but we have to print it somehow anyway...
            # XXX is it ok?
            length += 1

    else:
        mid = ext

    if length == 1:
        return c1

    res = []
    next = (length - 2)//2
    nmid = (length - 2) - next*2

    res += [top]
    res += [ext]*next
    res += [mid]*nmid
    res += [ext]*next
    res += [bot]

    return res


def vobj(symb, height):
    """Construct vertical object of a given height.

    See Also
    ========

    xobj
    """
    return '\n'.join( xobj(symb, height) )


def hobj(symb, width):
    """Construct horizontal object of a given width.

    See Also
    ========

    xobj
    """
    return ''.join( xobj(symb, width) )


# RADICAL
# n -> symbol
root = {
    2: '\N{SQUARE ROOT}',   # '\N{RADICAL SYMBOL BOTTOM}'
    3: '\N{CUBE ROOT}',
    4: '\N{FOURTH ROOT}',
}


# RATIONAL
def VF(txt):
    return U(f'VULGAR FRACTION {txt}')


# (p,q) -> symbol
frac = {
    (1, 2): VF('ONE HALF'),
    (1, 3): VF('ONE THIRD'),
    (2, 3): VF('TWO THIRDS'),
    (1, 4): VF('ONE QUARTER'),
    (3, 4): VF('THREE QUARTERS'),
    (1, 5): VF('ONE FIFTH'),
    (2, 5): VF('TWO FIFTHS'),
    (3, 5): VF('THREE FIFTHS'),
    (4, 5): VF('FOUR FIFTHS'),
    (1, 6): VF('ONE SIXTH'),
    (5, 6): VF('FIVE SIXTHS'),
    (1, 8): VF('ONE EIGHTH'),
    (3, 8): VF('THREE EIGHTHS'),
    (5, 8): VF('FIVE EIGHTHS'),
    (7, 8): VF('SEVEN EIGHTHS'),
}


# atom symbols
_xsym = {
    '==':  ('=', '='),
    '<':   ('<', '<'),
    '>':   ('>', '>'),
    '<=':  ('<=', '\N{LESS-THAN OR EQUAL TO}'),
    '>=':  ('>=', '\N{GREATER-THAN OR EQUAL TO}'),
    '!=':  ('!=', '\N{NOT EQUAL TO}'),
    '*':   ('*', '\N{DOT OPERATOR}'),
    '-->': ('-->', '\N{EM DASH}' + '\N{EM DASH}' +
            '\N{BLACK RIGHT-POINTING TRIANGLE}' if '\N{EM DASH}'
            and '\N{BLACK RIGHT-POINTING TRIANGLE}' else None),
    '==>': ('==>', '\N{BOX DRAWINGS DOUBLE HORIZONTAL}' +
            '\N{BOX DRAWINGS DOUBLE HORIZONTAL}' +
            '\N{BLACK RIGHT-POINTING TRIANGLE}' if
            '\N{BOX DRAWINGS DOUBLE HORIZONTAL}' and
            '\N{BOX DRAWINGS DOUBLE HORIZONTAL}' and
            '\N{BLACK RIGHT-POINTING TRIANGLE}' else None),
    '.':   ('*', '\N{RING OPERATOR}'),
}


def xsym(sym):
    """Get symbology for a 'character'."""
    op = _xsym[sym]

    if _use_unicode:
        return op[1]
    else:
        return op[0]


# SYMBOLS

atoms_table = {
    # class                    how-to-display
    'Exp1':                    '\N{SCRIPT SMALL E}',
    'Pi':                      '\N{GREEK SMALL LETTER PI}',
    'Infinity':                '\N{INFINITY}',
    'NegativeInfinity':        '\N{INFINITY}' and ('-' + '\N{INFINITY}'),  # XXX what to do here
    # 'ImaginaryUnit':          '\N{GREEK SMALL LETTER IOTA}',
    # 'ImaginaryUnit':          '\N{MATHEMATICAL ITALIC SMALL I}',
    'ImaginaryUnit':           '\N{DOUBLE-STRUCK ITALIC SMALL I}',
    'EmptySet':                '\N{EMPTY SET}',
    'Naturals':                '\N{DOUBLE-STRUCK CAPITAL N}',
    'Naturals0':               ('\N{DOUBLE-STRUCK CAPITAL N}' and
                                ('\N{DOUBLE-STRUCK CAPITAL N}' +
                                 '\N{SUBSCRIPT ZERO}')),
    'Integers':                '\N{DOUBLE-STRUCK CAPITAL Z}',
    'Rationals':               '\N{DOUBLE-STRUCK CAPITAL Q}',
    'Reals':                   '\N{DOUBLE-STRUCK CAPITAL R}',
    'Union':                   '\N{UNION}',
    'SymmetricDifference':     '\N{INCREMENT}',
    'Intersection':            '\N{INTERSECTION}',
    'Ring':                    '\N{RING OPERATOR}'
}


def pretty_atom(atom_name, default=None):
    """Return pretty representation of an atom."""
    if _use_unicode:
        return atoms_table[atom_name]
    else:
        if default is not None:
            return default

        raise KeyError('only unicode')  # send it default printer


def pretty_symbol(symb_name):
    """Return pretty representation of a symbol."""
    # let's split symb_name into symbol + index
    # UC: beta1
    # UC: f_beta

    if not _use_unicode:
        return symb_name

    name, sups, subs = split_super_sub(symb_name)

    def translate(s):
        gG = greek_unicode.get(s)
        if gG is not None:
            return gG
        for key in sorted(modifier_dict, key=lambda k: len(k), reverse=True):
            if s.lower().endswith(key) and len(s) > len(key):
                return modifier_dict[key](translate(s[:-len(key)]))
        return s

    name = translate(name)

    # Let's prettify sups/subs. If it fails at one of them, pretty sups/subs are
    # not used at all.
    def pretty_list(l, mapping):
        result = []
        for s in l:
            pretty = mapping.get(s)
            if pretty is None:
                try:  # match by separate characters
                    pretty = ''.join([mapping[c] for c in s])
                except (TypeError, KeyError):
                    return
            result.append(pretty)
        return result

    pretty_sups = pretty_list(sups, sup)
    if pretty_sups is not None:
        pretty_subs = pretty_list(subs, sub)
    else:
        pretty_subs = None

    # glue the results into one string
    if pretty_subs is None:  # nice formatting of sups/subs did not work
        if subs:
            name += '_'+'_'.join([translate(s) for s in subs])
        if sups:
            name += '__'+'__'.join([translate(s) for s in sups])
        return name
    else:
        sups_result = ' '.join(pretty_sups)
        subs_result = ' '.join(pretty_subs)

    return ''.join([name, sups_result, subs_result])


def annotated(letter):
    """
    Return a stylised drawing of the letter ``letter``, together with
    information on how to put annotations (super- and subscripts to the
    left and to the right) on it.

    See pretty.py functions _print_meijerg, _print_hyper on how to use this
    information.
    """
    ucode_pics = {
        'F': (2, 0, 2, 0, '\N{BOX DRAWINGS LIGHT DOWN AND RIGHT}\N{BOX DRAWINGS LIGHT HORIZONTAL}\n'
                          '\N{BOX DRAWINGS LIGHT VERTICAL AND RIGHT}\N{BOX DRAWINGS LIGHT HORIZONTAL}\n'
                          '\N{BOX DRAWINGS LIGHT UP}'),
        'G': (3, 0, 3, 1,
              '\N{BOX DRAWINGS LIGHT ARC DOWN AND RIGHT}\N{BOX DRAWINGS LIGHT HORIZONTAL}\N{BOX DRAWINGS LIGHT ARC DOWN AND LEFT}\n'
              '\N{BOX DRAWINGS LIGHT VERTICAL}\N{BOX DRAWINGS LIGHT RIGHT}\N{BOX DRAWINGS LIGHT DOWN AND LEFT}\n'
              '\N{BOX DRAWINGS LIGHT ARC UP AND RIGHT}\N{BOX DRAWINGS LIGHT HORIZONTAL}\N{BOX DRAWINGS LIGHT ARC UP AND LEFT}')
    }
    ascii_pics = {
        'F': (3, 0, 3, 0, ' _\n|_\n|\n'),
        'G': (3, 0, 3, 1, ' __\n/__\n\\_|')
    }

    if _use_unicode:
        return ucode_pics[letter]
    else:
        return ascii_pics[letter]



from ...core import Expr, sympify
from .matexpr import MatrixExpr


class FunctionMatrix(MatrixExpr):
    """
    Represents a Matrix using a function (Lambda)

    This class is an alternative to SparseMatrix

    >>> i, j = symbols('i j')
    >>> X = FunctionMatrix(3, 3, Lambda((i, j), i + j))
    >>> Matrix(X)
    Matrix([
    [0, 1, 2],
    [1, 2, 3],
    [2, 3, 4]])

    >>> Y = FunctionMatrix(1000, 1000, Lambda((i, j), i + j))

    >>> isinstance(Y*Y, MatMul)  # this is an expression object
    True

    >>> (Y**2)[10, 10]  # So this is evaluated lazily
    342923500
    """

    def __new__(cls, rows, cols, lamda):
        rows, cols = sympify(rows), sympify(cols)
        return Expr.__new__(cls, rows, cols, lamda)

    @property
    def shape(self):
        return self.args[0:2]

    @property
    def lamda(self):
        return self.args[2]

    def _entry(self, i, j):
        return self.lamda(i, j)

    def _eval_trace(self):
        from ...concrete import Sum
        from .trace import Trace
        return Trace(self).rewrite(Sum)



import pytest

from diofant import (Identity, ImmutableMatrix, MatAdd, MatMul, MatPow,
                     MatrixSymbol, Rational, ShapeError, Sum, ZeroMatrix, pi,
                     symbols)


__all__ = ()

n, m, l, k = symbols('n m l k', integer=True)
A = MatrixSymbol('A', n, m)
B = MatrixSymbol('B', m, l)
C = MatrixSymbol('C', n, n)
D = MatrixSymbol('D', n, n)
E = MatrixSymbol('E', m, n)


def test_matpow():
    pytest.raises(TypeError, lambda: MatPow(1, 1))


def test_entry():
    assert MatPow(A, 1)[0, 0] == A[0, 0]
    assert MatPow(C, 0)[0, 0] == 1
    assert MatPow(C, 0)[0, 1] == 0
    assert isinstance(MatPow(C, 2)[0, 0], Sum)

    pytest.raises(NotImplementedError, lambda: MatPow(C, -1)[0, 0])


def test_as_explicit_symbol():
    X = MatrixSymbol('X', 2, 2)
    assert MatPow(X, 0).as_explicit() == ImmutableMatrix(Identity(2))
    assert MatPow(X, 1).as_explicit() == X.as_explicit()
    assert MatPow(X, 2).as_explicit() == (X.as_explicit())**2


def test_as_explicit_nonsquare_symbol():
    X = MatrixSymbol('X', 2, 3)
    assert MatPow(X, 1).as_explicit() == X.as_explicit()
    for r in [0, 2, Rational(1, 2), pi]:
        pytest.raises(ShapeError, lambda: MatPow(X, r).as_explicit())


def test_as_explicit():
    A = ImmutableMatrix([[1, 2], [3, 4]])
    assert MatPow(A, 0).as_explicit() == ImmutableMatrix(Identity(2))
    assert MatPow(A, 1).as_explicit() == A
    assert MatPow(A, 2).as_explicit() == A**2
    assert MatPow(A, -1).as_explicit() == A.inv()
    assert MatPow(A, -2).as_explicit() == (A.inv())**2
    # less expensive than testing on a 2x2
    A = ImmutableMatrix([4])
    assert MatPow(A, Rational(1, 2)).as_explicit() == A**Rational(1, 2)


def test_as_explicit_nonsquare():
    A = ImmutableMatrix([[1, 2, 3], [4, 5, 6]])
    assert MatPow(A, 1).as_explicit() == A
    pytest.raises(ShapeError, lambda: MatPow(A, 0).as_explicit())
    pytest.raises(ShapeError, lambda: MatPow(A, 2).as_explicit())
    pytest.raises(ShapeError, lambda: MatPow(A, -1).as_explicit())
    pytest.raises(ValueError, lambda: MatPow(A, pi).as_explicit())


def test_doit_nonsquare_MatrixSymbol():
    assert MatPow(A, 1).doit() == A
    for r in [0, 2, -1, pi]:
        assert MatPow(A, r).doit() == MatPow(A, r)


def test_doit_square_MatrixSymbol_symsize():
    assert MatPow(C, 0).doit() == Identity(n)
    assert MatPow(C, 0).doit(deep=False) == Identity(n)
    assert MatPow(C, 1).doit() == C
    for r in [2, -1, pi]:
        assert MatPow(C, r).doit() == MatPow(C, r)


def test_doit_with_MatrixBase():
    X = ImmutableMatrix([[1, 2], [3, 4]])
    assert MatPow(X, 0).doit() == ImmutableMatrix(Identity(2))
    assert MatPow(X, 1).doit() == X
    assert MatPow(X, 2).doit() == X**2
    assert MatPow(X, -1).doit() == X.inv()
    assert MatPow(X, -2).doit() == (X.inv())**2
    # less expensive than testing on a 2x2
    assert MatPow(ImmutableMatrix([4]), Rational(1, 2)).doit() == ImmutableMatrix([2])


def test_doit_nonsquare():
    X = ImmutableMatrix([[1, 2, 3], [4, 5, 6]])
    assert MatPow(X, 1).doit() == X
    pytest.raises(ShapeError, lambda: MatPow(X, 0).doit())
    pytest.raises(ShapeError, lambda: MatPow(X, 2).doit())
    pytest.raises(ShapeError, lambda: MatPow(X, -1).doit())
    pytest.raises(ShapeError, lambda: MatPow(X, pi).doit())


def test_doit_nested_MatrixExpr():
    X = ImmutableMatrix([[1, 2], [3, 4]])
    Y = ImmutableMatrix([[2, 3], [4, 5]])
    assert MatPow(MatMul(X, Y), 2).doit() == (X*Y)**2
    assert MatPow(MatAdd(X, Y), 2).doit() == (X + Y)**2


def test_identity_power():
    k = Identity(n)
    assert MatPow(k, 4).doit() == k
    assert MatPow(k, n).doit() == k
    assert MatPow(k, -3).doit() == k
    assert MatPow(k, 0).doit() == k
    l = Identity(3)
    assert MatPow(l, n).doit() == l
    assert MatPow(l, -1).doit() == l
    assert MatPow(l, 0).doit() == l


def test_zero_power():
    z1 = ZeroMatrix(n, n)
    assert MatPow(z1, 3).doit() == z1
    pytest.raises(ValueError, lambda: MatPow(z1, -1).doit())
    assert MatPow(z1, 0).doit() == Identity(n)
    assert MatPow(z1, n).doit() == z1
    pytest.raises(ValueError, lambda: MatPow(z1, -2).doit())
    z2 = ZeroMatrix(4, 4)
    assert MatPow(z2, n).doit() == z2
    pytest.raises(ValueError, lambda: MatPow(z2, -3).doit())
    assert MatPow(z2, 2).doit() == z2
    assert MatPow(z2, 0).doit() == Identity(4)
    pytest.raises(ValueError, lambda: MatPow(z2, -1).doit())



"""
Package for symbolic vector algebra in 3D.
"""

from .coordsysrect import CoordSysCartesian
from .deloperator import Del
from .dyadic import BaseDyadic, Dyadic, DyadicAdd, DyadicMul, DyadicZero
from .functions import (curl, divergence, express, gradient, is_conservative,
                        is_solenoidal, matrix_to_vector, scalar_potential,
                        scalar_potential_difference)
from .orienters import (AxisOrienter, BodyOrienter, QuaternionOrienter,
                        SpaceOrienter)
from .point import Point
from .scalar import BaseScalar
from .vector import BaseVector, Vector, VectorAdd, VectorMul, VectorZero


__all__ = ('CoordSysCartesian', 'Del', 'BaseDyadic', 'Dyadic', 'DyadicAdd',
           'DyadicMul', 'DyadicZero', 'curl', 'divergence', 'express',
           'gradient', 'is_conservative', 'is_solenoidal', 'matrix_to_vector',
           'scalar_potential', 'scalar_potential_difference', 'AxisOrienter',
           'BodyOrienter', 'QuaternionOrienter', 'SpaceOrienter', 'Point',
           'BaseScalar', 'BaseVector', 'Vector', 'VectorAdd',
           'VectorMul', 'VectorZero')



from diofant import Rational, Sum, exp, factorial
from diofant.abc import x, z
from diofant.stats import E, density, variance
from diofant.stats.drv_types import (Geometric, GeometricDistribution, Poisson,
                                     PoissonDistribution)


__all__ = ()


def test_PoissonDistribution():
    l = 3
    p = PoissonDistribution(l)
    assert abs(p.cdf(10).evalf() - 1) < .001
    assert p.expectation(x, x) == l
    assert p.expectation(x**2, x) - p.expectation(x, x)**2 == l


def test_Poisson():
    l = 3
    x = Poisson('x', l)
    assert E(x) == l
    assert variance(x) == l
    assert density(x) == PoissonDistribution(l)
    assert isinstance(E(x, evaluate=False), Sum)
    assert isinstance(E(2*x, evaluate=False), Sum)
    assert density(x)(z) == 3**z/(exp(3)*factorial(z))


def test_GeometricDistribution():
    p = Rational(1, 5)
    d = GeometricDistribution(p)
    assert d.expectation(x, x) == 1/p
    assert d.expectation(x**2, x) - d.expectation(x, x)**2 == (1-p)/p**2
    assert abs(d.cdf(20000).evalf() - 1) < .001

    X = Geometric('x', p)
    assert density(X)(z) == Rational(4, 5)**(z - 1)/5



from ..core import Mul
from ..core.compatibility import default_sort_key
from ..functions import DiracDelta, Heaviside


def change_mul(node, x):
    """change_mul(node, x)

    Rearranges the operands of a product, bringing to front any simple
    DiracDelta expression.

    If no simple DiracDelta expression was found, then all the DiracDelta
    expressions are simplified (using DiracDelta.simplify).

    Return: (dirac, new node)
    Where:
        o dirac is either a simple DiracDelta expression or None (if no simple
          expression was found);
        o new node is either a simplified DiracDelta expressions or None (if it
          could not be simplified).

    Examples
    ========

    >>> change_mul(x*y*DiracDelta(x)*cos(x), x)
    (DiracDelta(x), x*y*cos(x))
    >>> change_mul(x*y*DiracDelta(x**2 - 1)*cos(x), x)
    (None, x*y*cos(x)*DiracDelta(x - 1)/2 + x*y*cos(x)*DiracDelta(x + 1)/2)
    >>> change_mul(x*y*DiracDelta(cos(x))*cos(x), x)
    (None, None)

    See Also
    ========

    diofant.functions.special.delta_functions.DiracDelta
    deltaintegrate

    """
    if not (node.is_Mul or node.is_Pow):
        return node

    new_args = []
    dirac = None

    # Sorting is needed so that we consistently collapse the same delta;
    # However, we must preserve the ordering of non-commutative terms
    c, nc = node.args_cnc()
    sorted_args = sorted(c, key=default_sort_key)
    sorted_args.extend(nc)

    for arg in sorted_args:
        if arg.is_Pow and isinstance(arg.base, DiracDelta):
            new_args.append(arg.func(arg.base, arg.exp - 1))
            arg = arg.base
        if dirac is None and (isinstance(arg, DiracDelta) and arg.is_simple(x)
                              and (len(arg.args) <= 1 or arg.args[1] == 0)):
            dirac = arg
        else:
            new_args.append(arg)
    if not dirac:  # there was no simple dirac
        new_args = []
        for arg in sorted_args:
            if isinstance(arg, DiracDelta):
                new_args.append(arg.simplify(x))
            elif arg.is_Pow and isinstance(arg.base, DiracDelta):
                new_args.append(arg.func(arg.base.simplify(x), arg.exp))
            else:
                new_args.append(change_mul(arg, x))
        if new_args != sorted_args:
            nnode = Mul(*new_args).expand()
        else:  # if the node didn't change there is nothing to do
            nnode = None
        return None, nnode
    return dirac, Mul(*new_args)


def deltaintegrate(f, x):
    """
    deltaintegrate(f, x)

    The idea for integration is the following:

    - If we are dealing with a DiracDelta expression, i.e. DiracDelta(g(x)),
      we try to simplify it.

      If we could simplify it, then we integrate the resulting expression.
      We already know we can integrate a simplified expression, because only
      simple DiracDelta expressions are involved.

      If we couldn't simplify it, there are two cases:

      1) The expression is a simple expression: we return the integral,
         taking care if we are dealing with a Derivative or with a proper
         DiracDelta.

      2) The expression is not simple (i.e. DiracDelta(cos(x))): we can do
         nothing at all.

    - If the node is a multiplication node having a DiracDelta term:

      First we expand it.

      If the expansion did work, then we try to integrate the expansion.

      If not, we try to extract a simple DiracDelta term, then we have two
      cases:

      1) We have a simple DiracDelta term, so we return the integral.

      2) We didn't have a simple term, but we do have an expression with
         simplified DiracDelta terms, so we integrate this expression.

    Examples
    ========

        >>> deltaintegrate(x*sin(x)*cos(x)*DiracDelta(x - 1), x)
        sin(1)*cos(1)*Heaviside(x - 1)
        >>> deltaintegrate(y**2*DiracDelta(x - z)*DiracDelta(y - z), y)
        z**2*DiracDelta(x - z)*Heaviside(y - z)

    See Also
    ========

    diofant.functions.special.delta_functions.DiracDelta
    diofant.integrals.integrals.Integral

    """
    if not f.has(DiracDelta):
        return

    from .integrals import Integral, integrate
    from ..solvers import solve

    # g(x) = DiracDelta(h(x))
    if f.func == DiracDelta:
        h = f.simplify(x)
        if h == f:  # can't simplify the expression
            # FIXME: the second term tells whether is DeltaDirac or Derivative
            # For integrating derivatives of DiracDelta we need the chain rule
            if f.is_simple(x):
                if (len(f.args) <= 1 or f.args[1] == 0):
                    return Heaviside(f.args[0])
                else:
                    return (DiracDelta(f.args[0], f.args[1] - 1) /
                            f.args[0].as_poly().LC())
        else:  # let's try to integrate the simplified expression
            fh = integrate(h, x)
            return fh
    elif f.is_Mul or f.is_Pow:  # g(x) = a*b*c*f(DiracDelta(h(x)))*d*e
        g = f.expand()
        if f != g:  # the expansion worked
            fh = integrate(g, x)
            if fh is not None and not isinstance(fh, Integral):
                return fh
        else:
            # no expansion performed, try to extract a simple DiracDelta term
            dg, rest_mult = change_mul(f, x)

            if not dg:
                if rest_mult:
                    fh = integrate(rest_mult, x)
                    return fh
            else:
                dg = dg.simplify(x)
                if dg.is_Mul:  # Take out any extracted factors
                    dg, rest_mult_2 = change_mul(dg, x)
                    rest_mult = rest_mult*rest_mult_2
                point = solve(dg.args[0], x)[0][x]
                return (rest_mult.subs({x: point})*Heaviside(x - point))



"""This module is intended for solving recurrences (difference equations)."""

import collections

from ..concrete import product
from ..core import (Add, Dummy, Equality, Function, Integer, Lambda, Mul,
                    Rational, Symbol, Wild, oo, sympify)
from ..core.compatibility import default_sort_key, iterable
from ..functions import FallingFactorial, RisingFactorial, binomial, factorial
from ..matrices import Matrix, casoratian
from ..polys import Poly, gcd, gcd_list, lcm_list, quo, resultant, roots
from ..simplify import hypersimilar, hypersimp
from ..utilities import numbered_symbols
from .ode import constantsimp
from .solvers import solve


def rsolve_poly(coeffs, f, n):
    r"""
    Find polynomial solutions for linear recurrence.

    Given linear recurrence operator `\operatorname{L}` of order
    `k` with polynomial coefficients and inhomogeneous equation
    `\operatorname{L} y = f`, where `f` is a polynomial, we seek for
    all polynomial solutions over field `K` of characteristic zero.

    Notes
    =====

    The algorithm performs two basic steps:

        1. Compute degree `N` of the general polynomial solution.
        2. Find all polynomials of degree `N` or less
           of `\operatorname{L} y = f`.

    There are two methods for computing the polynomial solutions.
    If the degree bound is relatively small, i.e. it's smaller than
    or equal to the order of the recurrence, then naive method of
    undetermined coefficients is being used. This gives system
    of algebraic equations with `N+1` unknowns.

    In the other case, the algorithm performs transformation of the
    initial equation to an equivalent one, for which the system of
    algebraic equations has only `r` indeterminates. This method is
    quite sophisticated (in comparison with the naive one) and was
    invented together by Abramov, Bronstein and Petkovšek.

    It is possible to generalize the algorithm implemented here to
    the case of linear q-difference and differential equations.

    Examples
    ========

    Lets say that we would like to compute `m`-th Bernoulli polynomial
    up to a constant, using `b(n+1) - b(n) = m n^{m-1}` recurrence:

    >>> rsolve_poly([-1, 1], 4*n**3, n)
    (C0 + n**4 - 2*n**3 + n**2, [C0])
    >>> bernoulli(4, n)
    n**4 - 2*n**3 + n**2 - 1/30

    References
    ==========

    * :cite:`Abramov1995polynomial`
    * :cite:`Petkovsek1992hyper`
    * :cite:`Petkovsek1997AeqB`

    """
    f = sympify(f)

    if not f.is_polynomial(n):
        return

    homogeneous = f.is_zero

    r = len(coeffs) - 1

    coeffs = [Poly(coeff, n) for coeff in coeffs]

    g = gcd_list(coeffs + [f], n, polys=True)
    if not g.is_ground:
        coeffs = [quo(c, g, n, polys=False) for c in coeffs]
        f = quo(f, g, n, polys=False)

    polys = [Poly(0, n)] * (r + 1)
    terms = [(Integer(0), -oo)] * (r + 1)

    for i in range(r + 1):
        for j in range(i, r + 1):
            polys[i] += coeffs[j] * binomial(j, i)

        if not polys[i].is_zero:
            (exp,), coeff = polys[i].LT()
            terms[i] = (coeff, exp)

    d = b = terms[0][1]

    for i in range(1, r + 1):
        if terms[i][1] > d:
            d = terms[i][1]

        if terms[i][1] - i > b:
            b = terms[i][1] - i

    d, b = int(d), int(b)

    x = Dummy('x')

    degree_poly = Integer(0)

    for i in range(r + 1):
        if terms[i][1] - i == b:
            degree_poly += terms[i][0] * FallingFactorial(x, i)

    nni_roots = list(roots(degree_poly, x, filter='Z',
                           predicate=lambda r: r >= 0))

    if nni_roots:
        N = [max(nni_roots)]
    else:
        N = []

    if homogeneous:
        N += [-b - 1]
    else:
        N += [f.as_poly(n).degree() - b, -b - 1]

    N = int(max(N))

    if N < 0:
        if homogeneous:
            return Integer(0), []
        else:
            return

    if N <= r:
        C = []
        y = E = Integer(0)

        for i in range(N + 1):
            C.append(Symbol('C' + str(i)))
            y += C[i] * n**i

        for i in range(r + 1):
            E += coeffs[i].as_expr() * y.subs({n: n + i})

        solutions = solve((E - f).as_poly(n).coeffs(), *C)

        if solutions:
            solutions = solutions[0]

        C = [c for c in C if (c not in solutions)]
        result = y.subs(solutions)
    else:
        A = r
        U = N + A + b + 1

        nni_roots = list(roots(polys[r], filter='Z',
                               predicate=lambda r: r >= 0))

        if nni_roots != []:
            a = max(nni_roots) + 1
        else:
            a = Integer(0)

        def _zero_vector(k):
            return [Integer(0)] * k

        def _one_vector(k):
            return [Integer(1)] * k

        def _delta(p, k):
            B = Integer(1)
            D = p.subs({n: a + k})

            for i in range(1, k + 1):
                B *= -Rational(k - i + 1, i)
                D += B * p.subs({n: a + k - i})

            return D

        alpha = {}

        for i in range(-A, d + 1):
            E = _one_vector(d + 1)

            for k in range(1, d + 1):
                E[k] = E[k - 1] * (x + i - k + 1) / k

            alpha[i] = Integer(0)

            for j in range(A + 1):
                for k in range(d + 1):
                    B = binomial(k, i + j)
                    D = _delta(polys[j].as_expr(), k)

                    alpha[i] += E[k] * B * D

        V = Matrix(U, A, lambda i, j: int(i == j))

        if homogeneous:
            for i in range(A, U):
                v = _zero_vector(A)

                for k in range(1, A + b + 1):
                    if i - k < 0:
                        break

                    B = alpha[k - A].subs({x: i - k})

                    for j in range(A):
                        v[j] += B * V[i - k, j]

                denom = alpha[-A].subs({x: i})

                for j in range(A):
                    V[i, j] = -v[j] / denom
        else:
            G = _zero_vector(U)

            for i in range(A, U):
                v = _zero_vector(A)
                g = Integer(0)

                for k in range(1, A + b + 1):
                    if i - k < 0:
                        break

                    B = alpha[k - A].subs({x: i - k})

                    for j in range(A):
                        v[j] += B * V[i - k, j]

                    g += B * G[i - k]

                denom = alpha[-A].subs({x: i})

                for j in range(A):
                    V[i, j] = -v[j] / denom

                G[i] = (_delta(f, i - A) - g) / denom

        P, Q = _one_vector(U), _zero_vector(A)

        for i in range(1, U):
            P[i] = (P[i - 1] * (n - a - i + 1) / i).expand()

        for i in range(A):
            Q[i] = Add(*[(v * p).expand() for v, p in zip(V[:, i], P)])

        if not homogeneous:
            h = Add(*[(g * p).expand() for g, p in zip(G, P)])

        C = [Symbol('C' + str(i)) for i in range(A)]

        def g(i):
            return Add(*[c * _delta(q, i) for c, q in zip(C, Q)])

        if homogeneous:
            E = [g(i) for i in range(N + 1, U)]
        else:
            E = [g(i) + _delta(h, i) for i in range(N + 1, U)]

        if E != []:
            solutions = solve(E, *C)
            solutions = solutions[0]
        else:
            solutions = {}

        if homogeneous:
            result = Integer(0)
        else:
            result = h

        for c, q in list(zip(C, Q)):
            if c in solutions:
                s = solutions[c] * q
                C.remove(c)
            else:
                s = c * q

            result += s.expand()

    return result, C


def rsolve_ratio(coeffs, f, n):
    r"""
    Find rational solutions for linear recurrence.

    Given linear recurrence operator `\operatorname{L}` of order `k`
    with polynomial coefficients and inhomogeneous equation
    `\operatorname{L} y = f`, where `f` is a polynomial, we seek
    for all rational solutions over field `K` of characteristic zero.

    Notes
    =====

    The algorithm performs two basic steps:

        1. Compute polynomial `v(n)` which can be used as universal
           denominator of any rational solution of equation
           `\operatorname{L} y = f`.
        2. Construct new linear difference equation by substitution
           `y(n) = u(n)/v(n)` and solve it for `u(n)` finding all its
           polynomial solutions. Return :obj:`None` if none were found.

    Algorithm implemented here is a revised version of the original
    Abramov's algorithm, developed in 1989. The new approach is much
    simpler to implement and has better overall efficiency. This
    method can be easily adapted to q-difference equations case.

    Besides finding rational solutions alone, this functions is
    an important part of the Hyper algorithm were it is used to find
    particular solution of inhomogeneous part of a recurrence.

    Examples
    ========

    >>> rsolve_ratio([-2*n**3 + n**2 + 2*n - 1, 2*n**3 + n**2 - 6*n,
    ...               -2*n**3 - 11*n**2 - 18*n - 9,
    ...               2*n**3 + 13*n**2 + 22*n + 8], 0, n)
    (C2*(2*n - 3)/(2*(n**2 - 1)), [C2])

    References
    ==========

    * :cite:`Abramov1995rational`

    See Also
    ========

    rsolve_hyper

    """
    f = sympify(f)

    if not f.is_polynomial(n):
        return

    coeffs = list(map(sympify, coeffs))

    r = len(coeffs) - 1

    A, B = coeffs[r], coeffs[0]
    A = A.subs({n: n - r}).expand()

    h = Dummy('h')

    res = resultant(A, B.subs({n: n + h}), n)
    assert res.is_polynomial(n)

    nni_roots = list(roots(res, h, filter='Z',
                           predicate=lambda r: r >= 0))

    if not nni_roots:
        return rsolve_poly(coeffs, f, n)
    else:
        C, numers = Integer(1), [Integer(0)] * (r + 1)

        for i in range(max(nni_roots), -1, -1):
            d = gcd(A, B.subs({n: n + i}), n)

            A = quo(A, d, n)
            B = quo(B, d.subs({n: n - i}), n)

            C *= Mul(*[d.subs({n: n - j}) for j in range(i + 1)])

        denoms = [C.subs({n: n + i}) for i in range(r + 1)]

        for i in range(r + 1):
            g = gcd(coeffs[i], denoms[i], n)

            numers[i] = quo(coeffs[i], g, n)
            denoms[i] = quo(denoms[i], g, n)

        for i in range(r + 1):
            numers[i] *= Mul(*(denoms[:i] + denoms[i + 1:]))

        result = rsolve_poly(numers, f * Mul(*denoms), n)

        if result is not None:
            return (result[0] / C).simplify(), result[1]


def rsolve_hyper(coeffs, f, n):
    r"""
    Find hypergeometric solutions for linear recurrence.

    Given linear recurrence operator `\operatorname{L}` of order `k`
    with polynomial coefficients and inhomogeneous equation
    `\operatorname{L} y = f` we seek for all hypergeometric solutions
    over field `K` of characteristic zero.

    The inhomogeneous part can be either hypergeometric or a sum
    of a fixed number of pairwise dissimilar hypergeometric terms.

    Notes
    =====

    The algorithm performs three basic steps:

        1. Group together similar hypergeometric terms in the
           inhomogeneous part of `\operatorname{L} y = f`, and find
           particular solution using Abramov's algorithm.
        2. Compute generating set of `\operatorname{L}` and find basis
           in it, so that all solutions are linearly independent.
        3. Form final solution with the number of arbitrary
           constants equal to dimension of basis of `\operatorname{L}`.

    The output of this procedure is a linear combination of fixed
    number of hypergeometric terms.  However the underlying method
    can generate larger class of solutions - D'Alembertian terms.

    This method not only computes the kernel of the
    inhomogeneous equation, but also reduces in to a basis so that
    solutions generated by this procedure are linearly independent.

    Examples
    ========

    >>> rsolve_hyper([-1, 1], 1 + n, n)
    (C0 + n*(n + 1)/2, [C0])

    References
    ==========

    * :cite:`Petkovsek1992hyper`
    * :cite:`Petkovsek1997AeqB`

    """
    coeffs = list(map(sympify, coeffs))

    f = sympify(f)

    r, kernel, symbols = len(coeffs) - 1, [], set()

    if not f.is_zero:
        if f.is_Add:
            similar = {}

            for g in f.expand().args:
                if not g.is_hypergeometric(n):
                    return

                for h in list(similar):
                    if hypersimilar(g, h, n):
                        similar[h] += g
                        break
                else:
                    similar[g] = Integer(0)

            inhomogeneous = []

            for g, h in similar.items():
                inhomogeneous.append(g + h)
        elif f.is_hypergeometric(n):
            inhomogeneous = [f]
        else:
            return

        for i, g in enumerate(inhomogeneous):
            coeff, polys = Integer(1), coeffs[:]
            denoms = [Integer(1)] * (r + 1)

            g = g.simplify()
            s = hypersimp(g, n)

            for j in range(1, r + 1):
                coeff *= s.subs({n: n + j - 1})

                p, q = coeff.as_numer_denom()

                polys[j] *= p
                denoms[j] = q

            for j in range(r + 1):
                polys[j] *= Mul(*(denoms[:j] + denoms[j + 1:]))

            R = rsolve_ratio(polys, Mul(*denoms), n)
            if R is not None:
                R, syms = R
                if syms:
                    R = R.subs(zip(syms, [0] * len(syms)))

            if R:
                inhomogeneous[i] *= R
            else:
                return

            result = Add(*inhomogeneous)
            result = result.simplify()
    else:
        result = Integer(0)

    Z = Dummy('Z')

    p, q = coeffs[0], coeffs[r].subs({n: n - r + 1})

    p_factors = list(roots(p, n))
    q_factors = list(roots(q, n))

    factors = [(Integer(1), Integer(1))]

    for p in p_factors:
        for q in q_factors:
            if p.is_integer and q.is_integer and p <= q:
                continue
            else:
                factors += [(n - p, n - q)]

    p = [(n - p, Integer(1)) for p in p_factors]
    q = [(Integer(1), n - q) for q in q_factors]

    factors = p + factors + q

    for A, B in factors:
        polys, degrees = [], []
        D = A * B.subs({n: n + r - 1})

        for i in range(r + 1):
            a = Mul(*[A.subs({n: n + j}) for j in range(0, i)])
            b = Mul(*[B.subs({n: n + j}) for j in range(i, r)])

            poly = quo(coeffs[i] * a * b, D, n)
            polys.append(poly.as_poly(n))

            if not poly.is_zero:
                degrees.append(polys[i].degree())

        d, poly = max(degrees), Integer(0)

        for i in range(r + 1):
            coeff = polys[i].coeff_monomial((d,))

            if coeff != 0:
                poly += coeff * Z**i

        for z in roots(poly, Z):
            if z.is_zero:
                continue

            sol, syms = rsolve_poly([polys[i] * z**i for i in range(r + 1)],
                                    0, n)
            sol = sol.collect(syms)
            sol = [sol.coeff(_) for _ in syms]

            for C in sol:
                ratio = z * A * C.subs({n: n + 1}) / B / C
                ratio = ratio.simplify()

                skip = max([-1] + [v for v in roots(Mul(*ratio.as_numer_denom()), n)
                                   if v.is_Integer]) + 1
                K = product(ratio, (n, skip, n - 1))

                if K.has(factorial, FallingFactorial, RisingFactorial):
                    K = K.simplify()

                if casoratian(kernel + [K], n, zero=False) != 0:
                    kernel.append(K)

    kernel.sort(key=default_sort_key)
    sk = list(zip(numbered_symbols('C'), kernel))

    for C, ker in sk:
        result += C * ker

    symbols |= {s for s, k in sk}
    return result, sorted(symbols, key=default_sort_key)


def rsolve(f, *y, init={}, simplify=True):
    r"""
    Solve recurrence equations.

    The equations can involve objects of the form `y(n + k)`, where
    `k` is a constant.

    Parameters
    ==========

    f : Expr, Equality or iterable of above
        The single recurrence equation or a system of recurrence
        equations.

    \*y : tuple
        Holds function applications `y(n)`, wrt to which the recurrence
        equation(s) will be solved.  If none given (empty tuple), this
        will be guessed from the provided equation(s).

    init : dict, optional
        The initial/boundary conditions for the recurrence equations as
        mapping of the function application `y(n_i)` to its value.
        Default is empty dictionary.

    simplify : bool, optional
        Enable simplification (default) on solutions.

    Examples
    ========

    >>> eq = (n - 1)*f(n + 2) - (n**2 + 3*n - 2)*f(n + 1) + 2*n*(n + 1)*f(n)

    >>> rsolve(eq)
    [{f: Lambda(n, 2**n*C0 + C1*factorial(n))}]
    >>> rsolve(eq, init={f(0): 0, f(1): 3})
    [{f: Lambda(n, 3*2**n - 3*factorial(n))}]

    Notes
    =====

    Currently, the function can handle linear recurrences with polynomial
    coefficients and hypergeometric inhomogeneous part.

    See Also
    ========

    diofant.solvers.ode.dsolve : solving differential equations
    diofant.solvers.solvers.solve : solving algebraic equations

    """
    if not iterable(f):
        f = [f]

    f = [_.lhs - _.rhs if isinstance(_, Equality) else _ for _ in f]
    f = [_.expand() for _ in f]

    if len(f) > 1 or len(y) > 1:
        raise NotImplementedError('Support for systems of recurrence '
                                  'equations is not implemented yet.')
    else:
        f = f[0]

    if not y:
        y = sorted(f.atoms(Function), key=default_sort_key)[0]
    else:
        y = y[0]
    n = y.args[0]

    h_part = collections.defaultdict(lambda: Integer(0))
    i_part = Integer(0)

    for h, c in f.collect(y.func(Wild('n')), evaluate=False).items():
        if h.func == y.func:
            k = Wild('k', exclude=(n,))
            r = h.args[0].match(n + k)
            if r:
                c = c.simplify()
                if not c.is_rational_function(n):
                    raise ValueError(f"Rational function of '{n}' expected, got '{c}'")
                h_part[int(r[k])] = c
            else:
                raise ValueError(f"'{y.func}({n} + Integer)' expected, got '{h}'")
        else:
            i_term = h * c
            if i_term.find(y.func(Wild('k'))):
                raise NotImplementedError(f"Linear recurrence for '{y.func}' "
                                          f"expected, got '{f}'")
            i_part -= i_term

    if not i_part.is_zero:
        if not all(p.is_hypergeometric(n) for p in i_part.as_coeff_add(n)[1]):
            raise NotImplementedError('Inhomogeneous part should be a sum of '
                                      f"hypergeometric terms in '{n}', got "
                                      f"'{i_part}'")

    k_min, k_max = min(h_part), max(h_part)

    if k_min < 0:
        return rsolve(f.subs({n: n + abs(k_min)}), y, init=init,
                      simplify=simplify)

    i_numer, i_denom = i_part.as_numer_denom()

    common = lcm_list([x.as_numer_denom()[1] for x in h_part.values()] + [i_denom])

    if common != 1:
        for k, coeff in h_part.items():
            numer, denom = coeff.as_numer_denom()
            h_part[k] = numer * quo(common, denom, n)

        i_part = i_numer * quo(common, i_denom, n)

    coeffs = [h_part[i] for i in range(k_max + 1)]

    result = rsolve_hyper(coeffs, i_part, n)

    if result is None:
        return

    solution, symbols = result

    if symbols and init != {}:
        equations = []

        for k, v in init.items():
            if k.is_Function and k.func == y.func:
                i = int(k.args[0])
            else:
                raise ValueError(f"'{y.func}(Integer)' expected, got '{k}'")
            eq = solution.limit(n, i) - v
            equations.append(eq)

        result = solve(equations, *symbols)

        if not result:
            return
        else:
            solution = solution.subs(result[0])

    if simplify:
        solution = solution.expand(log=True, mul=False)
        solution = constantsimp(solution, symbols)
        solution = solution.simplify()

    return [{y.func: Lambda((n,), solution)}]
